\HELP THE FROG by Loren Blaney, 2010
\(Based on Amy's First Primer by Computing Specialties)
\
\Want to learn XPL0 programing language? Start by making this small change:
\Line 11: "Cols=10" -> "Cols=5"
\
\Don't forget to disable word wrap!

inc CODESI;                      \code definitions for intrinsic routines

def Cols=10, Rows=6;             \dimensions of maze (cells; max. = 10x6)

def SizeX=31, SizeY=28,          \maze cell dimensions (pixels)
    MazeX = (10-Cols)*SizeX/2+4, \offsets that center maze on the screen
    MazeY = (6-Rows)*SizeY/2+0;

def Bel=$07, Esc=$1B;            \control characters
def UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;   \key scan codes
def Black, Blue, Green, Cyan, Red, Magenta, Brown, White, \attribute colors
    Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \EGA palette

proc Delay(T); \Delay T eighteenths of a second (approximately)
    int T;
    Sound(false, T, 1);

proc HopSound(Rising); \Interesting sound effect
    int Rising; \flag: rising pitch
    int I, J;
    begin
        for I:= 0, 2 do
            begin
                Sound(false, 1, 1);
                J:= if Rising then 2-I else I;
                Sound(J, 1, 4000*J);
            end;
    end; \HopSound

proc DrawSprite(X0, Y0, S); \Draw a sprite
    int  X0, Y0; \coordinates of upper-left corner (pixels)
    char S;      \address of sprite data
    int  X, Y, I, C, W, H;
    begin
        W:= S(0); \get width and height (in pixels)
        H:= S(1);
        I:= 2;
        for Y:= 0, H-1 do
            for X:= 0, W-1 do
                begin
                    C:= S(I)>>4;  \high nibble is left-most pixel
                    if C#9 then Point(X0+X, Y0+Y, C!$80);
                    X:= X+1;
                    C:= S(I)&$0F; \get pixel's color
                    if C#9 then Point(X0+X, Y0+Y, C!$80);
                    I:= I+1;
                end;
    end; \DrawSprite

proc DrawWall(X0, Y0, X1, Y1); \Draw wall (or ceiling) of maze
    int X0, Y0, X1, Y1; \starting and ending coordinates of line
    def OffX=MazeX, OffY=MazeY;
    begin
        Move(X0*SizeX+OffX, Y0*SizeY+OffY);
        Line(X1*SizeX+OffX, Y1*SizeY+OffY, %00010001<<8+LRed);
        Move(X0*SizeX+OffX+1, Y0*SizeY+OffY+1);
        Line(X1*SizeX+OffX+1, Y1*SizeY+OffY+1, %01000100<<8+Yellow);
        Move(X0*SizeX+OffX+2, Y0*SizeY+OffY+2);
        Line(X1*SizeX+OffX+2, Y1*SizeY+OffY+2, %00010001<<8+LRed);
    end; \DrawWall

proc DrawPath(X0, Y0, X1, Y1); \Draw frog's pathway
    int X0, Y0, X1, Y1;
    def OffX=SizeX/2+MazeX-1, OffY=SizeY/2+MazeY+1;
    int T;
    begin
        if X1 < X0 then \end points in ascending order for dotted lines
            begin \to be erased when back going over the same path
                T:= X1;  X1:= X0;  X0:= T; \swap points
                T:= Y1;  Y1:= Y0;  Y0:= T;
            end;
        if Y1 < Y0 then \end points in ascending order for dotted lines
            begin
                T:= X1;  X1:= X0;  X0:= T; \swap points
                T:= Y1;  Y1:= Y0;  Y0:= T;
            end;
        Move(X0*SizeX+OffX, Y0*SizeY+OffY);
        Line (X1*SizeX+OffX, Y1*SizeY+OffY, %00110011<<8+LGreen+$80);
    end; \DrawPath

proc DrawFrog(X, Y); \Draw frog image
    int  X, Y;
    def  OffX=3+MazeX, OffY=3+MazeY;
    char S;
    begin
        S:= [28+25<<8, \28x25 nibble array = 7x25 word array
            $9999, $9999, $9999, $8888, $9899, $2222, $9989,
            $9999, $9999, $9899, $2222, $8229, $1121, $9928,
            $9999, $9999, $8299, $1121, $2222, $FF1F, $8912,
            $9999, $9999, $2298, $FF1F, $2112, $11FF, $29F1,
            $9999, $9999, $2192, $11FF, $21F1, $11FF, $2911,
            $9999, $9999, $2182, $11FF, $2111, $11FF, $2911,
            $9999, $9999, $2182, $11FF, $2211, $FF1F, $8912,
            $9999, $9999, $2288, $FF1F, $2212, $1121, $8922,
            $8998, $9899, $2282, $1121, $2222, $2222, $282A,
            $2898, $9899, $A222, $2222, $1A2A, $2212, $2822,
            $2282, $8889, $2A22, $1122, $2222, $2122, $2912,
            $A282, $8228, $2282, $2222, $1111, $1211, $8922,
            $AA2A, $2282, $8828, $2282, $2222, $2822, $2888,
            $282A, $2A22, $22AA, $8828, $8888, $AA88, $28AA,
            $2288, $A222, $2222, $AA22, $AAAA, $AAAA, $29AA,
            $2282, $222A, $A22A, $AAAA, $EEAE, $AAAA, $89A2,
            $2298, $2AA2, $2AA2, $AEAA, $EAEE, $A2AA, $282A,
            $2282, $2AA2, $AAA2, $EEAA, $EAEE, $22AA, $29AA,
            $2282, $2AA2, $AA22, $EEAE, $AAEE, $2222, $29A2,
            $2282, $2A22, $AA22, $AAAA, $A2AA, $2A22, $89A2,
            $8298, $2A22, $22A2, $A2AA, $2222, $AA22, $9922,
            $9899, $2282, $22A2, $2822, $8888, $A22A, $9928,
            $9999, $8298, $8822, $9988, $9899, $282A, $9999,
            $9999, $8299, $2222, $9999, $9299, $8922, $9999,
            $9999, $9899, $2828, $9999, $2899, $2928, $9999];
        DrawSprite(X*SizeX+OffX, Y*SizeY+OffY, S);
    end; \DrawFrog

proc DrawPond(X, Y); \Draw lily pond image
    int  X, Y;
    def  OffX=3+MazeX, OffY=6+MazeY;
    char S;
    begin
        S:= [28+20<<8, \28x20 nibble array = 7x20 word array
            $9999, $9999, $9999, $3393, $9399, $9999, $9999,
            $9999, $9399, $9933, $9999, $9999, $9999, $9933,
            $9999, $9999, $9999, $9999, $9399, $9933, $9999,
            $3399, $9939, $3399, $3333, $9933, $9999, $3399,
            $9999, $3399, $9933, $9999, $9999, $7377, $9933,
            $3393, $9939, $9999, $3399, $3333, $9929, $9999,
            $9999, $3399, $3333, $3733, $337E, $9923, $3399,
            $9933, $9999, $AA9A, $AAAA, $DFFF, $332A, $9333,
            $3399, $AA9A, $22AA, $2222, $F22F, $AA2A, $99AA,
            $3A99, $AAAA, $2222, $2222, $2222, $2222, $A32A,
            $AA9A, $A2AA, $2222, $2222, $2222, $2222, $AA22,
            $AA9A, $F2AA, $22F2, $2222, $2222, $2222, $AA22,
            $9A99, $FAFA, $F2D2, $2A22, $22A2, $2A22, $A3AA,
            $3999, $DE9F, $22DF, $2222, $AAAA, $A3AA, $9933,
            $3333, $EDFE, $FAEE, $AAAA, $AAAA, $39AA, $9999,
            $9999, $FF99, $33E9, $9999, $3393, $9939, $3399,
            $9933, $779F, $99F7, $3393, $9933, $9999, $9999,
            $3399, $9939, $3393, $3933, $3399, $3333, $9999,
            $9999, $9999, $9399, $9933, $9999, $9999, $3933,
            $9933, $9999, $9999, $3399, $9933, $9999, $9999];
        DrawSprite(X*SizeX+OffX, Y*SizeY+OffY, S);
    end; \DrawPond

int  FlyX, FlyY; \coordinates of upper-left corner of fly sprite
char FlyS;       \address of fly sprite image data

proc DrawFly(X, Y); \Draw tasty fly
    int X, Y;
    def OffX=MazeX+8, OffY=MazeY+5;
    int I;
    begin
        FlyS:= [20+12<<8, \20x12 nibble array = 5x12 word array
            $9999, $9999, $9999, $7799, $9977,
            $7797, $9979, $9991, $9977, $7979,
            $7979, $7997, $C9C8, $9779, $7999,
            $9779, $7199, $8187, $7999, $9987,
            $8997, $9897, $7875, $8787, $9979,
            $7799, $8887, $7875, $7998, $9997,
            $9997, $9878, $1817, $9779, $9997,
            $9797, $7199, $3133, $7797, $9979,
            $7799, $9877, $3833, $9999, $9999,
            $9999, $9199, $1113, $9999, $9999,
            $9999, $9999, $8981, $9999, $9999,
            $9999, $9999, $9999, $9999, $9999];

        FlyX:= X*SizeX+OffX;
        FlyY:= Y*SizeY+OffY;
        DrawSprite(FlyX, FlyY, FlyS); \draw fly
        for I:= 1, 18 do \make fly buzz around awhile...
            begin
                Delay(2);
                DrawSprite(FlyX, FlyY, FlyS); \erase fly
                FlyX:= FlyX+Ran(3)-1;
                FlyY:= FlyY+Ran(3)-1;
                DrawSprite(FlyX, FlyY, FlyS); \draw fly
            end;
    end; \DrawFly

proc DrawTongue; \Draw frog's Tongue zapping the fly
    begin
        Move((Cols-2)*SizeX+MazeX+23, Rows*SizeY+MazeY+14);
        Line(FlyX+5, FlyY+4, LRed+$80);
    end; \DrawTongue

\============================================================================

int Key; \keystroke value (ASCII or negative scan code)

proc GetKey; \Wait for, then return keystroke's value
    begin
        Key:= ChIn(1);
        if Key = Esc then [SetVid($03); exit]; \exit program?
        if Key=0 then Key:= -Chin(1);          \get scan code 
    end; \GetKey

int Cell(Cols+1, Rows+1, 3);      \cells (plus right and bottom borders)
def LeftWall, Ceiling, Connected; \attributes of each cell (= 0, 1 and 2)

proc ConnectFrom(X, Y); \Connect maze cells starting from cell X,Y
    int X, Y;
    int Dir, Dir0;
    begin
        Cell(X, Y, Connected):= true; \mark current cell as connected
        Dir:= Ran(4);                 \randomly choose a direction
        Dir0:= Dir;                   \save this initial direction
        repeat
            case Dir of \try to connect to cell at Dir
                0: if X+1<Cols & not Cell(X+1, Y, Connected) then \go right
                    [Cell(X+1, Y, LeftWall):= false; ConnectFrom(X+1, Y)];
                1: if Y+1<Rows & not Cell(X, Y+1, Connected) then \go down
                    [Cell(X, Y+1, Ceiling):= false; ConnectFrom(X, Y+1)];
                2: if X-1>=0 & not Cell(X-1, Y, Connected) then   \go left
                    [Cell(X, Y, LeftWall):= false; ConnectFrom(X-1, Y)];
                3: if Y-1>=0 & not Cell(X, Y-1, Connected) then   \go up
                    [Cell(X, Y, Ceiling):= false; ConnectFrom(X, Y-1)]
                other []; \(never occurs)
            Dir:= Dir+1 & $03; \next direction
        until Dir = Dir0;
    end; \ConnectFrom

proc PlayGame; \Help the frog find his way out of the maze
    int X, Y, Moves;
    begin
        for Y:= 0, Rows do \generate a random maze
            for X:= 0, Cols do
                begin
                    Cell(X, Y, LeftWall):= true;   \start with all walls and
                    Cell(X, Y, Ceiling):= true;    \ceilings in place
                    Cell(X, Y, Connected):= false; \and cells disconnected
                end;
        Cell(Cols-2, Rows, Ceiling):= false; \make exit door, w. room for fly

        ConnectFrom(Ran(Cols), Ran(Rows));   \randomly pick a starting cell

        SetVid($0D);     \320x200 graphics in 16 colors
        port($3C8):= 0;  \set background to light blue
        port($3C9):= 21; \red
        port($3C9):= 21; \green
        port($3C9):= 63; \blue

        for Y:= 0, Rows do \display the maze
            begin
                for X:= 0, Cols do
                if X#Cols & Cell(X, Y, Ceiling) then DrawWall(X, Y, X+1, Y);
                for X:= 0, Cols do
                if Y#Rows & Cell(X, Y, LeftWall) then DrawWall(X, Y, X, Y+1);
            end;

        Cursor(1, 22);
        Attrib(Yellow);
        Text(6, "Help the frog find the pond.");
        Cursor(2, 23); ChOut(6, $18); \show arrow keys
        Cursor(1, 24); ChOut(6, $1B); ChOut(6, $19); ChOut(6, $1A);

        DrawPond(Cols-2, Rows);

        Moves:= 0;     \reset move counter
        X:= 0;  Y:= 0; \start frog at upper doorway
        repeat
			DrawFrog(X, Y);
			OpenI(1);       \flush type-ahead buffer
            HopSound(true);
            GetKey;
            DrawFrog(X, Y); \erase frog (by XOR drawing)
            case Key of
                -UpArrow: if Cell(X, Y, Ceiling) = false then
                    begin
                        DrawPath(X, Y, X, Y-1);
                        Y:= Y-1;
                        Moves:= Moves+1; \count this move
                    end;
                -DnArrow: if Cell(X, Y+1, Ceiling) = false then
                    begin
                        DrawPath(X, Y, X, Y+1);
                        Y:= Y+1;
                    end;
                -LtArrow: if Cell(X, Y, LeftWall) = false then
                    begin
                        DrawPath(X-1, Y, X, Y);
                        X:= X-1;
                    end;
                -RtArrow: if Cell(X+1, Y, LeftWall) = false then
                    begin
                        DrawPath(X+1, Y, X, Y);
                        X:= X+1;
                    end
            other ChOut(0, Bel); \wrong key!

            Moves:= Moves+1; \count this move
            Cursor(1+9, 24); \show the move count
            IntOut(6, Moves);
            if Moves # 1 then Text(6, " tries") else Text(6, " try");

        until X=Cols-2 & Y=Rows; \reached pond?
    end; \PlayGame

begin \Main
    loop
        begin
            PlayGame;
            DrawPond(Cols-2, Rows);       \erase pond by XOR drawing
            DrawFrog(Cols-2, Rows);       \draw frog in its place
            HopSound(false);
            DrawFly(Cols-1, Rows);        \draw fly buzzing around

            DrawTongue;                   \Zot!
            HopSound(true);
            DrawTongue;                   \erase Tongue
            DrawSprite(FlyX, FlyY, FlyS); \bye bye fly

            GetKey;                       \keystroke restarts game
        end; \Esc exits program
end; \Main
