\MEMORY (CONCENTRATION by Loren Blaney, 2010)
\
\Want to learn XPL0 programing language? Start by making this small change:
\Line 161: Blue -> Red
\
\Don't forget to disable word wrap!

inc CODESI;   \include code definitions for intrinsic routines

int CpuReg,   \address of CPU register array (from GetReg)
    Cards;    \number of cards remaining to be matched
int Deck(52); \deck of cards, contains values 0..51

def Black, Blue, Green, Cyan, Red, Magenta, Brown, White, \attribute colors
    Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite;

func CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
    int Int, AX, BX, CX, DX, BP, DS, ES; \(unused args need not be passed)
        begin
            CpuReg(0):= AX;
            CpuReg(1):= BX;
            CpuReg(2):= CX;
            CpuReg(3):= DX;
            CpuReg(6):= BP;
            CpuReg(9):= DS;
            CpuReg(11):= ES;
            SoftInt(Int);
            return CpuReg(0); \return AX register
        end; \CallInt

proc Blink(On); \Turn video blinking off and on
    \When blink is off, bright background colors are available
    int On; \flag: true = blink on; false = blink off
    CallInt($10, $1003, if On then 1 else 0);

\----------------------------- MOUSE ROUTINES -------------------------------

func OpenMouse; \Initializes mouse; returns 'false' if it fails
    begin \Pointer is set to center of screen but is hidden
        CallInt($21, $3533); \Be sure mouse vector ($33) points to something
        if CpuReg(1)=0 & CpuReg(11)=0 then return false;
        return Extend(CallInt($33, $0000)); \reset mouse and return status
    end; \OpenMouse (Extend is necessary for 32-bit XPL)

proc ShowMouse(On); \Turn mouse pointer on or off
    int On; \flag: true = pointer on; false = pointer off
    CallInt($33, if On then $0001 else $0002);

func GetMousePosition(N); \Return position of specified mouse coordinate
    int N; \0 = X coordinate; 1 = Y coordinate
    \For 80-column text modes divide the mouse coordinates by 8 to get the
    \character cursor position.
    begin
        CallInt($33, $0003);
        return if N then CpuReg(3) else CpuReg(2);
    end; \GetMousePosition

func GetMouseButton(N); \Return 'true' if specified mouse button is down
    int N; \button number: 0 = left; 1 = right (or middle)
    begin
        CallInt($33, $0003);
        return if N then (CpuReg(1)&2)=2 else (CpuReg(1)&1)=1;
    end; \GetMouseButton

\============================================================================

char TimeOfDay(4), \current time (0=hours, 1=mins, 2=secs, 3=huns)
     Timer(4),     \time since start of game
     Time0(4);     \initial time (sort of)
int  SecOld;       \used to only update display when seconds change

proc GetTimeOfDay; \Return time of day in TimeOfDay(4)
    begin
        CpuReg(0):= $2C00;
        SoftInt($21);
        TimeOfDay(0):= CpuReg(2)>>8; \hour
        TimeOfDay(1):= CpuReg(2);    \minute
        TimeOfDay(2):= CpuReg(3)>>8; \second
        TimeOfDay(3):= CpuReg(3);    \hundreths
    end; \GetTimeOfDay

proc ShowHMS(T); \Display hours : minutes : seconds
    char T; \time array (0=hours, 1=mins, 2=secs, 3=huns)
    int  I;
    begin
        ShowMouse(false);
        Cursor(35, 13);
        Attrib(BWhite<<4 + Black);

        for I:= 0, 2 do
            begin
                if T(I) < 10 then ChOut(6, ^0); \show leading zero
                IntOut(6, T(I));
                if I<2 then ChOut(6, ^:);
            end;

        ShowMouse(true);
    end; \ShowHMS

proc ElapsedTime(A, B, C); \Return C:= A-B [hrs(0), min(1), sec(2), hun(3)]
    char A, B, C;
    int  Borrow;
    begin
        C(3):= A(3) - B(3);          \hundreths
        Borrow:= 0;
        if Extend(C(3)) < 0 then [Borrow:= 1;  C(3):= C(3)+100];

        C(2):= A(2) - B(2) - Borrow; \seconds
        Borrow:= 0;
        if Extend(C(2)) < 0 then [Borrow:= 1;  C(2):= C(2)+60];

        C(1):= A(1) - B(1) - Borrow; \minutes
        Borrow:= 0;
        if Extend(C(1)) < 0 then [Borrow:= 1;  C(1):= C(1)+60];

        C(0):= A(0) - B(0) - Borrow; \hours
        if Extend(C(0)) < 0 then C(0):= C(0)+24;
    end; \ElapsedTime

proc TickTock; \Display seconds ticking away
    begin
        GetTimeOfDay;
        ElapsedTime(TimeOfDay, Time0, Timer);
        if Timer(2) # SecOld then \seconds have changed
            begin
                SecOld:= Timer(2);
                ShowHMS(Timer);
            end;
    end; \TickTock

\============================================================================

proc Exit; \Make a clean exit
    begin
        ShowMouse(false); \make sure mouse pointer is off
        Blink(true);      \(bright = false)
        Cursor(79, 23);   \so DOS prompt ends up at the bottom
        OpenI(0);         \Remove keypress from buffer
        exit;
    end; \Exit

proc ShowCard(X, Y, Card); \Show card C at coordinates X,Y
    int  X, Y, Card;
    int  N, Suit;
    char Str;

    proc CardStock(Ch); \Draw the basic card stock
        int Ch; \pattern for front or back side
        int I, J;
        begin
            for J:= 0, 2 do
                begin
                    Cursor(X, Y+J);
                    for I:= 0, 4 do ChOut(6, Ch);
                end;
        end; \CardStock

    begin
        if Card < 0 then \display the back of the card
            begin
                Attrib(BWhite<<4 + Blue);
                CardStock($B1);
                return;
            end;

        Str:= "234567891JQKA ";
        N:= Card/4;
        Suit:= Rem(0);

        Attrib(BWhite<<4 + (if Suit<2 then LRed else Black));
        CardStock(^ );

        Cursor(X+2, Y+1); \show suit symbol in center of card
        ChOut(6, Suit+3);

        Cursor(X, Y);     \show face value (number)
        ChOut(6, Str(N));
        if N = 10-2 then
            begin
                ChOut(6, ^0);
                Cursor(X+3, Y+2);
                Text(6, "10");
            end
        else
            begin
                Cursor(X+4, Y+2);
                ChOut(6, Str(N));
            end;
    end; \ShowCard

\----------------------------------------------------------------------------

proc Play; \Play the game
    int MX, MY,       \mouse coordinates (char cells)
        N, X, Y,      \card number and coordinates
        C1, C1X, C1Y, \first selected card
        C2, C2X, C2Y; \second selected card

func GetCard;
    \Given mouse coordinates MX,MY, return card N at position X,Y
    begin
        N:= 0;
        for Y:= 0, 5 do
            for X:= 0, 9 do
                begin
                    if Y#2 & Y#3 ! (X<3!X>6) then
                    begin
                        if MX<5+X*7+6 & MY<1+Y*4+4 then return;
                        N:= N+1;
                    end;
                end;
        N:= 52-1; \don't go beyond the last card
        X:= 9;  Y:= 5;
    end; \GetCard

proc PickCard; \Click on a card and turn it over
    begin
        ShowMouse(true);
        while GetMouseButton(0) do TickTock; \wait for button release
        repeat
            if ChkKey then Exit; \any keystroke terminates program
            TickTock;
        until GetMouseButton(0);

        MX:= GetMousePosition(0) / 8; \convert mouse x/y to character cells
        MY:= GetMousePosition(1) / 8;
        GetCard;
        ShowMouse(false);
        if Deck(N) >= 0 then ShowCard(5+X*7, 1+Y*4, Deck(N)); \card face up
    end; \PickCard

begin \Play
    PickCard;
    C1:= N;
    C1X:= X;
    C1Y:= Y;

    PickCard;
    C2:= N;
    C2X:= X;
    C2Y:= Y;

    if Deck(C1)/4 = Deck(C2)/4 & C1 # C2 & Deck(C1) >= 0 & Deck(C2) >= 0 then
        begin                \face values match and they aren't the same card
            Cards:= Cards-2; \and they have not already been taken
            Deck(C1):= -1;   \mark cards as taken
            Deck(C2):= -1;
        end
    else
        begin                \cards don't match (or illegal move)
            Sound(0, 36, 1); \delay then turn cards face down
            if Deck(C1) >= 0 then ShowCard(5+C1X*7, 1+C1Y*4, -1);
            if Deck(C2) >= 0 then ShowCard(5+C2X*7, 1+C2Y*4, -1);
        end;
end; \Play

\----------------------------------------------------------------------------

proc Deal; \Display all the cards facing down
    int X, Y;
    begin
        for Y:= 0, 5 do
            for X:= 0, 9 do
                if Y#2 & Y#3 ! (X<3!X>6) then ShowCard(5+X*7, 1+Y*4, -1);
    end; \Deal

proc Shuffle; \Shuffle the card deck
    int I, J, K, T;
    begin
        for J:= 0, 10-1 do
            for I:= 0, 52-1 do
                begin
                    K:= Ran(52); \randomly pick a card
                    T:= Deck(I); \swap cards in the deck at locations I and K
                    Deck(I):= Deck(K);
                    Deck(K):= T;
                end;
    end; \Shuffle

proc Init; \Initialize
    int  I, J;
    char Str;
    begin
        for I:= 0, 52-1 do Deck(I):= I; \set up the deck
        Shuffle;

        Cards:= 52;

        Attrib(Green<<4 + White);       \paint the background green
        for J:= 0, 25-1 do
            begin
                Cursor(0, J);
                for I:= 0, 80-2 do Chout(6, ^ );
            end;

        Str:= "MEMORY ";
        for I:= 0, 6-1 do
            begin
                Cursor(31+I+I+I, 10);
                Attrib(Green<<4 + ((I+4)&$0F));
                ChOut(6, Str(I));
            end;

        GetTimeOfDay; \get starting time
        for I:= 0, 3 do Time0(I):= TimeOfDay(I);
        SecOld:= -1;  \display timer immediately
    end; \Init

\============================================================================

begin \Main
    CpuReg:= GetReg;

    if not OpenMouse then
        begin
            Text(0, "This program requires a mouse");
            CrLf(0);
            exit;
        end;

    Blink(false);
    Init;
    Deal;

    repeat Play until Cards <= 0;

    \Show winning screen before exiting
    Blink(true);                   \(bright = false)
    while GetMouseButton(0) do []; \wait for button release
    repeat until GetMouseButton(0) ! ChkKey;
    Exit;
end; \Main
