\LOAD THE TRUCK by Loren Blaney, 2010
\(Based on Amy's First Primer by Computing Specialties)
\
\Want to learn XPL0 programing language? Start by making this small change:
\Line 298: true -> false
\
\Don't forget to disable word wrap!

inc CODESI; \code definitions for intrinsic routines

int Snd;    \flag: enable sound

def Bel=$07, Esc=$1B;                                     \control characters
def UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;   \key scan codes
def Black, Blue, Green, Cyan, Red, Magenta, Brown, White, \attribute colors
    Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite;

func GetKey; \Wait for, then return a keystroke's value
    int Key; \keystroke value (ASCII or negative scan code)
    begin
        Key:= ChIn(1);
        if Key = Esc then [SetVid($03); exit]; \exit program?
        if Key = ^s then Snd:= ~Snd;
        if Key=0 then Key:= -Chin(1);          \get scan code
        return Key;
end; \GetKey

proc Delay(T); \Delay T 18ths of a second (approximately)
    int T;
    Sound(false, T, 1);

proc DrawRectangle(X, Y, W, H, C); \Draw an unfilled rectangle
    int X, Y, W, H, \upper left corner coordinates, width, height (pixels)
        C;          \color
    begin
        Move(X, Y);
        Line(X+W-1, Y, C);
        Line(X+W-1, Y+H-1, C);
        Line(X, Y+H-1, C);
        Line(X, Y, C);
    end; \DrawRectangle

proc DrawSprite(X0, Y0, S); \Draw a sprite
    int  X0, Y0; \coordinates of upper-left corner (pixels)
    char S;      \address of sprite data
    int  X, Y, I, C, W, H;
    begin
        W:= S(0); \get width and height (in pixels)
        H:= S(1);
        I:= 2;
        for Y:= 0 to H-1 do
            for X:= 0 to W-1 do
                begin
                    C:= S(I)>>4;  \high nibble is left-most pixel
                    if C#0 & X0+X<320 then Point(X0+X, Y0+Y, C!$80);
                    X:= X+1;
                    C:= S(I)&$0F; \get pixel's color
                    if C#0 & X0+X<320 then Point(X0+X, Y0+Y, C!$80);
                    I:= I+1;
                end;
end; \DrawSprite

proc GetSprite(X0, Y0, X1, Y1, S); \Capture a sprite from the screen
    int  X0, Y0, X1, Y1; \upper-left and lower-right corners of sprite
    char S;              \WARNING: corners must be in ascending order
    int  X, Y, I, C1, C2;
    begin
        S(0):= X1-X0+1; \WARNING: width must be an even number
        S(1):= Y1-Y0+1; \height
        I:= 2;
        for Y:= Y0 to Y1 do
            for X:= X0 to X1 do
                begin
                    C1:= ReadPix(X, Y); \left pixel is high nibble
                    X:= X+1;
                    C2:= ReadPix(X, Y); \right pixel is low nibble
                    S(I):= C1<<4 ! (C2&$0F);
                    I:= I+1;
                end;
end; \GetSprite

\============================================================================

char Loader(2000), Trailer(2000), Cab(2000), Combo(4000), \sprite images
     Square(200), Diamond(200), Rectangle(200), Triangle(200), Circle(200);

proc DrawShape(X, Y, S); \Draw Shape S at X, Y
    int X, Y, S;
    begin
        case S of
            0: DrawSprite(X, Y, Square);
            1: DrawSprite(X, Y, Diamond);
            2: DrawSprite(X, Y, Rectangle);
            3: DrawSprite(X, Y, Triangle);
            4: DrawSprite(X, Y, Circle)
        other [];
    end; \DrawShape

proc PlayGame; \Load the truck
    int XL, YL,   \loader coordinates (pixels)
        XS, YS,   \shape (to be loaded) coordinates (pixels)
        XT, YT,   \truck (trailer) coordinates (pixels)
        Shape,    \shape to be loaded
        I, X, Y,  \scratch index and coordinates
        NLeft(5), \number of shapes left in each pile
        Order(4), \list of shapes to be loaded (and in that order)
        OrderInx;

proc LoadShapes; \Load the shapes onto the trailer
    loop begin
        Shape:= Order(OrderInx);
        YS:= Shape*25+15; \location of shape to pick up (XS,YS)
        I:= NLeft(Shape);
        XS:= I*15 - 15;

        \Let the loader drive around the screen until the Shape is found
        repeat OpenI(1); \flush type-ahead buffer
            repeat Sound(Snd, 1, 32162);
            until  ChkKey;
            DrawSprite(XL, YL, Loader); \erase loader (by XOR drawing)
            case GetKey of
                -UpArrow: if YL >=  5 then YL:= YL-5;
                -DnArrow: if YL < 150 then YL:= YL+5;
                -LtArrow: if XL >=  5 then XL:= XL-5;
                -RtArrow: if XL < 240 then XL:= XL+5
            other [];                   \wrong key!
            DrawSprite(XL, YL, Loader); \draw loader at new location

            \Quit if all shapes are picked and the loader is out of the way
            if OrderInx>=4 & YL<100 then quit;
        until XL=XS & YL=YS-5; \loop until a shape is found

        \The loader is at the shape, so raise its hoist
        GetSprite(XS, YS, XS+17, YS+15, Combo); \combine the shape and hoist
        for I:= 0 to 5-1 do
            begin
                DrawSprite(XS, YS, Combo); \erase images at old position
                YS:= YS-1;                 \raise hoist and shape
                DrawSprite(XS, YS, Combo); \draw them in the new position
                Sound(Snd, 1, 11333-456*I);
            end;
        Delay(2);

        \Combine the loader and shape into a single sprite
        GetSprite(XL, YL, XL+77, YL+25, Combo);

        \Let the loader holding a shape drive around until it gets to trailer
        XS:=110; YS:=140; \target location at front of trailer
        repeat
            begin
                OpenI(1); \flush type-ahead buffer
                repeat Sound(Snd, 1, 20877);
                until ChkKey;
                DrawSprite(XL, YL, Combo); \erase loader (by XOR drawing)
                case GetKey of
                    -UpArrow: if YL >=  5 then YL:= YL-5;
                    -DnArrow: if YL < 150 then YL:= YL+5;
                    -LtArrow: if XL >=  5 then XL:= XL-5;
                    -RtArrow: if XL < 240 then XL:= XL+5
                other [];                  \wrong key!
                DrawSprite(XL, YL, Combo); \draw loader at new location
            end;
        until XL=XS & YL=YS;

        \Loader is at the front of the trailer, so slide the shape onto it
        GetSprite(XS-46, YS, XS+17, YS+13, Combo); \include previous shapes
        for I:= 0 to 15-1 do
            begin
                DrawSprite(XS-46, YS, Combo); \erase at old position
                XS:= XS-1;
                DrawSprite(XS-46, YS, Combo); \draw in new position
                Sound(Snd, 1, 7438-376*I);
            end;
        Delay(2);

        \Retract the slider
        XS:= 110+1;
        GetSprite(XS, YS, XS+1, YS+13, Combo);
        for I:= 0 to 15-1 do
            begin
                DrawSprite(XS, YS, Combo); \erase at old position
                XS:= XS+1;
                DrawSprite(XS, YS, Combo); \draw in new position
                Sound(Snd, 1, 2174+376*I);
            end;
        Delay(2);

        \Lower the hoist
        XS:= XL+2;  YS:= YL;
        GetSprite(XS, YS, XS+15, YS+15, Combo);
        for I:= 0 to 5-1 do
            begin
                DrawSprite(XS, YS, Combo); \erase at old position
                YS:= YS+1;
                DrawSprite(XS, YS, Combo); \draw in new position
                Sound(Snd, 1, 9917+456*I);
            end;
        Delay(2);

        \Loop back to get the next shape
        Shape:= Order(OrderInx);
        NLeft(Shape):= NLeft(Shape)-1; \one less shape in the pile
        OrderInx:= OrderInx+1;         \point to next shape in the order box
    end; \LoadShapes; loop

\----------------------------------------------------------------------------

begin \PlayGame
    Clear; \erase screen
    Attrib(Yellow); Cursor(6, 24); Text(6, "L O A D   T H E   T R U C K");
    Attrib(LBlue);
    Cursor(38, 23); ChOut(6, $18); \show arrow keys
    Cursor(37, 24); ChOut(6, $1B); ChOut(6, $19); ChOut(6, $1A);

    \Fill the order box with four randomly selected shapes
    DrawRectangle(225, 13, 79, 19, White);
    for I:= 0 to 4-1 do
        begin
            Shape:= Ran(5);                 \randomly choose a shape
            DrawShape(233+I*18, 16, Shape); \draw it in the box
            Order(I):= Shape;               \and record it in the Order array
        end;

        for Y:= 0 to 5-1 do \draw the piles of shapes
            begin
                for X:= 0 to 5-1 do
                    DrawShape(X*15+4, Y*25+15+2, Y);
                NLeft(Y):= 5; \number of shapes left in each pile
            end;

        DrawSprite( 10, 152, Trailer); \display the rest of the sprites
        DrawSprite(247, 135, Cab);

        XL:= 145;  YL:= 50;   \put the loader at its starting position
        DrawSprite(XL, YL, Loader);
        OrderInx:= 0;         \set index to point to first shape in Order box

        LoadShapes;           \go load the ordered shapes

        \Trailer is fully loaded, loader is out of the way, so back the cab
        XT:= 247;   YT:= 135; \initial truck cab coordinates
        while XT > 100 do
            begin
                DrawSprite(XT, YT, Cab); \erase
                XT:= XT-5;
                DrawSprite(XT, YT, Cab); \draw
                Sound(Snd, 2, 11000);
            end;

        \Drive the cab with its trailer off the right edge of the screen
        Sound(Snd, 6, 4542*3); \honk, honk!
        Sound(0, 6, 1);
        Sound(Snd, 6, 4542*3);
        XT:= 10;  YT:=135;
        GetSprite(XT, YT, XT+173, 167, Combo);
        while XT < 320 do
            begin
                DrawSprite(XT, YT, Combo); \erase
                XT:= XT+5;
                DrawSprite(XT, YT, Combo); \draw
                Sound(Snd, 2, 22000);
            end;
        Delay(18*3);
    end; \PlayGame

\----------------------------------------------------------------------------

proc LoadBmp; \Load the TRUCK.BMP file onto the screen
    def W=320, H=200; \width and height of image
    int Hand,         \file handle
        I, X, Y;      \index and coordinates
    begin
        Trap(false);                  \don't trap "file not found" error
        Hand:= FOpen("TRUCK.BMP", 0); \open file for input
        if GetErr then [SetVid(3); Text(0, "TRUCK.BMP not found"); exit];
        Trap(true);                   \restore normal error trapping
        FSet(Hand, ^I);               \set device 3 to our handle
        OpenI(3);

        for I:= 0 to 53 do X:= ChIn(3);     \skip unused header info
        for I:= 0 to 16*4-1 do X:= ChIn(3); \skip color table

        for Y:= H-1 downto 0 do \.bmp files are upside down!
            for X:= 0 to W-1 do
                begin
                    I:= ChIn(3);
                    Point(X, Y, I>>4); \high nibble is left pixel
                    X:= X+1;
                    Point(X, Y, I);    \low nibble is right pixel
                end;

        FClose(Hand); \for properness
    end; \LoadBmp

\----------------------------------------------------------------------------

begin \Main
    Snd:= true;    \turn sound effects on

    SetVid($0D);   \set 320x200 graphics in 16 colors, erase screen
    LoadBmp;       \get our images

    \Set up the sprites from the figures on the screen
    GetSprite(  0,   7,  15,  19, Square);
    GetSprite(  0,  32,  15,  44, Diamond);
    GetSprite(  0,  57,  15,  69, Rectangle);
    GetSprite(  0,  82,  15,  94, Triangle);
    GetSprite(  0, 107,  15, 119, Circle);
    GetSprite(145,  50, 222,  76, Loader);
    GetSprite( 10, 152, 111, 167, Trailer);
    GetSprite(247, 135, 302, 167, Cab);

    loop PlayGame; \play until the Esc key is pressed
end; \Main
